#include <stdio.h>
#include <mem.h>
#include <malloc.h>

/*
Simple pager for QNX 4.25

Data:
	1. In buffer - buffer, which used for the reading file.
	2. Out buffer - string accumulator for the output.
	3. Inbuf size - size of the current input.
	4. Outbuf size - size of the current output.
	5. Line positions buffer - buffer, which contain line positions
		in the file.
	6. Line positions buffer size.
	7. Symbol counter in the current string.
	8. Quit flag.

Alg:
	1. While not quit flag:
		1. If not eof, read in buffer.
		2. While not in buffer end:
			1. While (out buffer string count < DISPLAY_LINES)
				and (out buffer pos < in buffer size)
				and (not in buffer end):
				1. Symbol counter = 1.
				2. While not string end found or in buffer end,
					read symblos in out buffer and increment symbol counter.
				3. Save file position in the line positions buffer (function).
			2. If (out buffer string count >= DISPLAY_LINES)
				or eof of out buffer size > max. buffer size:
				1. Display out buffer.
				2. Wait for keypress.
				3. If key == q, set quit flag.
				4. String counter = 0.
	2. Close file, free buffers and quit.

	Save file position function (t_line_buffer linebuf, pos):
		struct t_lpos_buffer
		{
			// Size of buffer (max elements).
			long buf_sz;
			// Count of lines in buffer.
			long line_count;
			// Pointer to the buffer.
			long *buf;
		} linebuf;

		1. If buffer not allocated, allocate space for the LB_BSZ elements.
		2. If line_count == buf_sz, realloc buffer: add LB_BSZ elements.
		3. buf[line_count++] = pos.

Conditions:
	1. String end - '\n' or LINE_MAX_WIDTH symbols.
	2. DISPLAY_LINES = 23 (Screen height = 25, minus 2 lines).

Artiom N.(cl)2012

*/

#ifdef __USAGE
%C - simple pager, like 'more' command.
Usage: pg <file>

Artiom N.(cl) 2012
#endif

//===========================================================================
#define DISPLAY_LINES 24
#define LINE_MAX_WIDTH 80
// Block for the alloc/realloc file positions buffer.
#define LB_BSZ 512
//===========================================================================
typedef struct t_lpos_buffer
{
	// Size of buffer (max elements).
	long buf_sz;
	// Count of lines in buffer.
	long line_count;
	// Pointer to the buffer.
	long *buf;
};
//---------------------------------------------------------------------------
typedef struct t_key
{
	char mod;
	char key;
};
//---------------------------------------------------------------------------
typedef int bool;
const bool false	= 0;
const bool true		= 1;
//---------------------------------------------------------------------------
typedef int t_retval;
//---------------------------------------------------------------------------
#define MAX_BUF_SZ DISPLAY_LINES * LINE_MAX_WIDTH
typedef struct t_buffer
{
	long buf_sz;
	long buf_pos;
	char buf[MAX_BUF_SZ];
};
//===========================================================================
bool save_file_pos(struct t_lpos_buffer *buf, const long pos)
{
	if (buf->line_count >= buf->buf_sz)
	{
		if (!(buf->buf = realloc(buf->buf, LB_BSZ))) return(false);
	}


	buf->buf[buf->line_count++] = pos;
	return(true);
}
//---------------------------------------------------------------------------
t_retval push_to_buf(struct t_buffer buf, const char sym)
{
	return(0);	
}
//---------------------------------------------------------------------------
char pop_from_buf(struct t_buffer buf)
{
	return(0);
}
//---------------------------------------------------------------------------
struct t_key get_key()
// Arrow keys have two key event.
{
	struct t_key key;
	key.mod = 0;
	key.key = getch();
	if (key.key == 0xff)
	{
		key.mod = 0xff;
		key.key = getch();
	}
	return(key);
}
//---------------------------------------------------------------------------
long read_string(struct t_buffer *src, struct t_buffer *dst)
// Read string from src to dest.
// Return count of symbols.
{
	register long sym_counter = 0;
	while ((++sym_counter < LINE_MAX_WIDTH) && (src->buf_pos < src->buf_sz))
	{
		(dst->buf[dst->buf_pos++] = src->buf[src->buf_pos++]);
		if (src->buf[src->buf_pos] == '\n') break;
	}

	return(sym_counter);
}
//---------------------------------------------------------------------------
int main(int argc, char **argv)
{
	struct t_lpos_buffer linebuf;
	struct t_buffer in_buf, out_buf;

	t_retval result = 0;
	bool quit_flag = false;
	register long ln_counter = 1;

	struct t_key ui_key;

	FILE *in_file, *out_file;

	memset(&linebuf, NULL, sizeof(linebuf));
	if (!(linebuf.buf = malloc(LB_BSZ)))
	{
		perror("Can't allocate memory!\nExiting...\n");
		result		= 1;
		quit_flag	= true;
	}
	linebuf.buf_sz = LB_BSZ;

	in_file = fopen(argv[1], "rb");

	if (in_file == NULL)
	{
		fprintf(stderr, "Can't open %\s\nExiting...\n", argv[1]);
		result		= 1;
		quit_flag	= true;
	}

	out_buf.buf_pos = out_buf.buf_sz = 0;

	out_file = stdout;

	while (!quit_flag)
	{
		if (!feof(in_file))
		{
			in_buf.buf_sz	= fread(in_buf.buf, sizeof(char), MAX_BUF_SZ, in_file);
			in_buf.buf_pos	= 0;
			for (; in_buf.buf_sz < MAX_BUF_SZ - 1; in_buf.buf_sz++)
				in_buf.buf[in_buf.buf_sz] = '\n';
		}
		else quit_flag = true;

		while ((in_buf.buf_pos < in_buf.buf_sz) && !quit_flag)
		{
			// 1. Reading in output buffer.
			while (
				(ln_counter++ < DISPLAY_LINES) &&
				(in_buf.buf_pos < in_buf.buf_sz) &&
				(out_buf.buf_pos < in_buf.buf_sz)
			)
			{
				read_string(&in_buf, &out_buf);
				if (!save_file_pos(&linebuf, ftell(in_file)))
				{
					perror("Can't allocate memory!\n");
					result		= 1;
					quit_flag	= true;
					break;
				};
			} // while ln_counter < DISPLAY_LINES

			out_buf.buf_sz = out_buf.buf_pos;


			// 2. Check and output.
			if (
				(ln_counter >= DISPLAY_LINES) ||
				(out_buf.buf_sz >= MAX_BUF_SZ) ||
				feof(in_file)
			)
			{

				fwrite(out_buf.buf, sizeof(char), out_buf.buf_sz, out_file);
				printf("\n ------- Press any key to continue or 'q' for the exit ------- ");
				fflush(out_file);
				ui_key = get_key();
				printf("\n\n");
				switch(ui_key.key)
				{
					case 'q':
						quit_flag = true;
						break;
					default: ;
				}
				ln_counter = 0;
				out_buf.buf_sz = out_buf.buf_pos = 0;
			}

		} // while not in_buf end
	} // while !eof
	
	fclose(in_file);
	free(linebuf.buf);
//	fclose(out_file);
	printf("\n");
	return(result);
}
//===========================================================================
